@@ -24,12 +24,17 @@
 #include "fluid_rev.h"
 #include "fluid_chorus.h"
 #include "fluidsynth_priv.h"
+#include "fluid_ladspa.h"
 
 #define SYNTH_REVERB_CHANNEL 0
 #define SYNTH_CHORUS_CHANNEL 1
 
 #define ENABLE_MIXER_THREADS 1
 
+// If less than x voices, the thread overhead is larger than the gain,
+// so don't activate the thread(s).
+#define VOICES_PER_THREAD 8
+
 typedef struct _fluid_mixer_buffers_t fluid_mixer_buffers_t;
 
 struct _fluid_mixer_buffers_t {
@@ -76,6 +81,10 @@ struct _fluid_rvoice_mixer_t {
   int active_voices; /**< Read-only: Number of non-null voices */
   int current_blockcount;      /**< Read-only: how many blocks to process this time */
 
+#ifdef LADSPA
+  fluid_LADSPA_FxUnit_t* LADSPA_FxUnit; /**< Used by mixer only: Effects unit for LADSPA support. Never created or freed */
+#endif
+
 #ifdef ENABLE_MIXER_THREADS
 //  int sleeping_threads;        /**< Atomic: number of threads currently asleep */
 //  int active_threads;          /**< Atomic: number of threads in the thread loop */
@@ -131,6 +140,38 @@ fluid_rvoice_mixer_process_fx(fluid_rvoice_mixer_t* mixer)
     }
     fluid_profile(FLUID_PROF_ONE_BLOCK_CHORUS, prof_ref);
   }
+  
+#ifdef LADSPA
+  /* Run the signal through the LADSPA Fx unit */
+  if (mixer->LADSPA_FxUnit) {
+    int j;
+    FLUID_DECLARE_VLA(fluid_real_t*, left_buf, mixer->buffers.buf_count);
+    FLUID_DECLARE_VLA(fluid_real_t*, right_buf, mixer->buffers.buf_count);
+    FLUID_DECLARE_VLA(fluid_real_t*, fx_left_buf, mixer->buffers.fx_buf_count);
+    FLUID_DECLARE_VLA(fluid_real_t*, fx_right_buf, mixer->buffers.fx_buf_count);
+    for (j=0; j < mixer->buffers.buf_count; j++) {
+      left_buf[j] = mixer->buffers.left_buf[j];
+      right_buf[j] = mixer->buffers.right_buf[j];
+    }
+    for (j=0; j < mixer->buffers.fx_buf_count; j++) {
+      fx_left_buf[j] = mixer->buffers.fx_left_buf[j];
+      fx_right_buf[j] = mixer->buffers.fx_right_buf[j];
+    }
+    for (i=0; i < mixer->current_blockcount * FLUID_BUFSIZE; i += FLUID_BUFSIZE) {
+      fluid_LADSPA_run(mixer->LADSPA_FxUnit, left_buf, right_buf, fx_left_buf, 
+		       fx_right_buf);
+      for (j=0; j < mixer->buffers.buf_count; j++) {
+        left_buf[j] += FLUID_BUFSIZE;
+        right_buf[j] += FLUID_BUFSIZE;
+      }
+      for (j=0; j < mixer->buffers.fx_buf_count; j++) {
+        fx_left_buf[j] += FLUID_BUFSIZE;
+        fx_right_buf[j] += FLUID_BUFSIZE;
+      }
+    }
+    fluid_check_fpe("LADSPA");
+  }
+#endif
 }
 
 /**
@@ -598,6 +639,14 @@ void delete_fluid_rvoice_mixer(fluid_rvoice_mixer_t* mixer)
 }
 
 
+#ifdef LADSPA				    
+void fluid_rvoice_mixer_set_ladspa(fluid_rvoice_mixer_t* mixer, 
+				   fluid_LADSPA_FxUnit_t* ladspa)
+{
+  mixer->LADSPA_FxUnit = ladspa;
+}
+#endif
+
 void fluid_rvoice_mixer_set_reverb_enabled(fluid_rvoice_mixer_t* mixer, int on)
 {
   mixer->fx.with_reverb = on;
@@ -655,6 +704,7 @@ fluid_mixer_get_mt_rvoice(fluid_rvoice_mixer_t* mixer)
 #define THREAD_BUF_PROCESSING 0
 #define THREAD_BUF_VALID 1
 #define THREAD_BUF_NODATA 2
+#define THREAD_BUF_TERMINATE 3
 
 /* Core thread function (processes voices in parallel to primary synthesis thread) */
 static void
@@ -676,8 +726,12 @@ fluid_mixer_thread_func (void* data)
       fluid_cond_mutex_unlock(mixer->thread_ready_m);
       
       fluid_cond_mutex_lock(mixer->wakeup_threads_m);
-      if (fluid_atomic_int_get(&buffers->ready) != THREAD_BUF_PROCESSING)
+      while (1) {
+        int j = fluid_atomic_int_get(&buffers->ready); 
+        if (j == THREAD_BUF_PROCESSING || j == THREAD_BUF_TERMINATE)
+          break;
         fluid_cond_wait(mixer->wakeup_threads, mixer->wakeup_threads_m);
+      }
       fluid_cond_mutex_unlock(mixer->wakeup_threads_m);
       
       hasValidData = 0;
@@ -729,13 +783,13 @@ fluid_mixer_buffers_mix(fluid_mixer_buffers_t* dest, fluid_mixer_buffers_t* src)
  * Go through all threads and see if someone is finished for mixing 
  */
 static FLUID_INLINE int
-fluid_mixer_mix_in(fluid_rvoice_mixer_t* mixer)
+fluid_mixer_mix_in(fluid_rvoice_mixer_t* mixer, int extra_threads)
 {
   int i, result, hasmixed;
   do {
     hasmixed = 0;
     result = 0;
-    for (i=0; i < mixer->thread_count; i++) {
+    for (i=0; i < extra_threads; i++) {
       int j = fluid_atomic_int_get(&mixer->threads[i].ready);
       switch (j) {
 	case THREAD_BUF_PROCESSING: 
@@ -755,23 +809,33 @@ fluid_mixer_mix_in(fluid_rvoice_mixer_t* mixer)
 static void 
 fluid_render_loop_multithread(fluid_rvoice_mixer_t* mixer)
 {
-  int i; //, test=0, waits=0;
+  int i, bufcount;
   //int scount = mixer->current_blockcount * FLUID_BUFSIZE;
   FLUID_DECLARE_VLA(fluid_real_t*, bufs, 
 		    mixer->buffers.buf_count * 2 + mixer->buffers.fx_buf_count * 2);
-  int bufcount = fluid_mixer_buffers_prepare(&mixer->buffers, bufs);
+  // How many threads should we start this time?
+  int extra_threads = mixer->active_voices / VOICES_PER_THREAD;
+  if (extra_threads > mixer->thread_count)
+    extra_threads = mixer->thread_count;
+  if (extra_threads == 0) {
+    // No extra threads? No thread overhead!
+    fluid_render_loop_singlethread(mixer);
+    return;
+  }
+
+  bufcount = fluid_mixer_buffers_prepare(&mixer->buffers, bufs);
   
   // Prepare voice list
   fluid_cond_mutex_lock(mixer->wakeup_threads_m);
   fluid_atomic_int_set(&mixer->current_rvoice, 0);
-  for (i=0; i < mixer->thread_count && i < (mixer->active_voices-1); i++)
+  for (i=0; i < extra_threads; i++)
     fluid_atomic_int_set(&mixer->threads[i].ready, THREAD_BUF_PROCESSING);
   // Signal threads to wake up
   fluid_cond_broadcast(mixer->wakeup_threads);
   fluid_cond_mutex_unlock(mixer->wakeup_threads_m);
   
   // If thread is finished, mix it in
-  while (fluid_mixer_mix_in(mixer)) {
+  while (fluid_mixer_mix_in(mixer, extra_threads)) {
     // Otherwise get a voice and render it
     fluid_rvoice_t* rvoice = fluid_mixer_get_mt_rvoice(mixer);
     if (rvoice != NULL) {
@@ -785,7 +849,7 @@ fluid_render_loop_multithread(fluid_rvoice_mixer_t* mixer)
       int is_processing = 0;
       //waits++;
       fluid_cond_mutex_lock(mixer->thread_ready_m);
-      for (i=0; i < mixer->thread_count; i++) 
+      for (i=0; i < extra_threads; i++) 
 	if (fluid_atomic_int_get(&mixer->threads[i].ready) == 
 	    THREAD_BUF_PROCESSING)
 	  is_processing = 1;
@@ -817,6 +881,8 @@ fluid_rvoice_mixer_set_threads(fluid_rvoice_mixer_t* mixer, int thread_count,
     fluid_atomic_int_set(&mixer->threads_should_terminate, 1);
     // Signal threads to wake up
     fluid_cond_mutex_lock(mixer->wakeup_threads_m);
+    for (i=0; i < mixer->thread_count; i++)
+      fluid_atomic_int_set(&mixer->threads[i].ready, THREAD_BUF_TERMINATE);
     fluid_cond_broadcast(mixer->wakeup_threads);
     fluid_cond_mutex_unlock(mixer->wakeup_threads_m);
   
